'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

exports.GomokuDirectionType = void 0;
(function (GomokuDirectionType) {
    GomokuDirectionType[GomokuDirectionType["LEFT"] = 0] = "LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP_LEFT"] = 2] = "TOP_LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP"] = 4] = "TOP";
    GomokuDirectionType[GomokuDirectionType["TOP_RIGHT"] = 6] = "TOP_RIGHT";
    GomokuDirectionType[GomokuDirectionType["RIGHT"] = 1] = "RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_RIGHT"] = 3] = "BOTTOM_RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM"] = 5] = "BOTTOM";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_LEFT"] = 7] = "BOTTOM_LEFT";
})(exports.GomokuDirectionType || (exports.GomokuDirectionType = {}));
const gomokuDirections = Array.from(Object.entries({
    [exports.GomokuDirectionType.TOP]: [-1, 0],
    [exports.GomokuDirectionType.TOP_RIGHT]: [-1, 1],
    [exports.GomokuDirectionType.RIGHT]: [0, 1],
    [exports.GomokuDirectionType.BOTTOM_RIGHT]: [1, 1],
    [exports.GomokuDirectionType.BOTTOM]: [1, 0],
    [exports.GomokuDirectionType.BOTTOM_LEFT]: [1, -1],
    [exports.GomokuDirectionType.LEFT]: [0, -1],
    [exports.GomokuDirectionType.TOP_LEFT]: [-1, -1],
}).reduce((acc, [key, value]) => {
    acc[key] = value;
    return acc;
}, []));
const leftHalfGomokuDirectionTypes = [
    exports.GomokuDirectionType.LEFT,
    exports.GomokuDirectionType.TOP_LEFT,
    exports.GomokuDirectionType.TOP,
    exports.GomokuDirectionType.TOP_RIGHT,
];
const rightHalfGomokuDirectionTypes = [
    exports.GomokuDirectionType.RIGHT,
    exports.GomokuDirectionType.BOTTOM_RIGHT,
    exports.GomokuDirectionType.BOTTOM,
    exports.GomokuDirectionType.BOTTOM_LEFT,
];
const gomokuDirectionTypes = leftHalfGomokuDirectionTypes.concat(rightHalfGomokuDirectionTypes);

class GomokuContext {
    MAX_ROW;
    MAX_COL;
    MAX_INLINE;
    NEXT_MOVER_MAX_BUFFER;
    TOTAL_POS;
    TOTAL_PLAYERS;
    gomokuDirections;
    idxMap;
    idxMaxMoveMap;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE, NEXT_MOVER_MAX_BUFFER = 0.4) {
        const _TOTAL_POS = MAX_ROW * MAX_COL;
        const _TOTAL_PLAYERS = 2;
        const _NEXT_MOVER_MAX_BUFFER = Math.max(0.1, Math.min(0.9, NEXT_MOVER_MAX_BUFFER));
        const _gomokuDirections = new Int32Array(gomokuDirections.map(([dr, dc]) => dr * MAX_ROW + dc));
        const _idxMap = new Array(_TOTAL_POS);
        for (let r = 0; r < MAX_ROW; ++r) {
            for (let c = 0; c < MAX_COL; ++c) {
                const id = r * MAX_ROW + c;
                _idxMap[id] = [r, c];
            }
        }
        const _idxMaxMoveMap = new Array(gomokuDirectionTypes.length);
        for (const dirType of gomokuDirectionTypes) {
            const maxMoveMap = new Int32Array(_TOTAL_POS);
            const [dr, dc] = gomokuDirections[dirType];
            for (let r = 0; r < MAX_ROW; ++r) {
                for (let c = 0; c < MAX_COL; ++c) {
                    let steps = 0;
                    for (let r2 = r, c2 = c;; ++steps) {
                        r2 += dr;
                        c2 += dc;
                        if (r2 < 0 || r2 >= MAX_ROW || c2 < 0 || c2 >= MAX_COL)
                            break;
                    }
                    const id = r * MAX_ROW + c;
                    maxMoveMap[id] = steps;
                }
            }
            _idxMaxMoveMap[dirType] = maxMoveMap;
        }
        this.MAX_ROW = MAX_ROW;
        this.MAX_COL = MAX_COL;
        this.MAX_INLINE = MAX_INLINE;
        this.TOTAL_POS = _TOTAL_POS;
        this.TOTAL_PLAYERS = _TOTAL_PLAYERS;
        this.NEXT_MOVER_MAX_BUFFER = _NEXT_MOVER_MAX_BUFFER;
        this.gomokuDirections = _gomokuDirections;
        this.idxMap = _idxMap;
        this.idxMaxMoveMap = _idxMaxMoveMap;
    }
    idx(r, c) {
        return r * this.MAX_ROW + c;
    }
    idxIfValid(r, c) {
        return this.isValidPos(r, c) ? r * this.MAX_ROW + c : -1;
    }
    revIdx(id) {
        return this.idxMap[id];
    }
    isValidPos(r, c) {
        return r >= 0 && r < this.MAX_ROW && c >= 0 && c < this.MAX_COL;
    }
    isInvalidPos(r, c) {
        return r < 0 || r >= this.MAX_ROW || c < 0 || c >= this.MAX_COL;
    }
    safeMove(id, dirType, step) {
        return step <= this.idxMaxMoveMap[dirType][id] ? id + this.gomokuDirections[dirType] * step : -1;
    }
    safeMoveOneStep(id, dirType) {
        return 1 <= this.idxMaxMoveMap[dirType][id] ? id + this.gomokuDirections[dirType] : -1;
    }
    fastMove(id, dirType, step) {
        return id + this.gomokuDirections[dirType] * step;
    }
    fastMoveOneStep(id, dirType) {
        return id + this.gomokuDirections[dirType];
    }
    maxMovableSteps(id, dirType) {
        return this.idxMaxMoveMap[dirType][id];
    }
    *validNeighbors(id) {
        for (const dirType of gomokuDirectionTypes) {
            const id2 = this.safeMoveOneStep(id, dirType);
            if (id2 >= 0)
                yield [id2, dirType];
        }
    }
    traverseAllDirections(handle) {
        this.traverseLeftDirections(handle);
        this.traverseRightDirections(handle);
    }
    traverseLeftDirections(handle) {
        const { TOTAL_POS } = this;
        for (const dirType of leftHalfGomokuDirectionTypes) {
            for (let id = 0; id < TOTAL_POS; ++id)
                handle(id, dirType);
        }
    }
    traverseRightDirections(handle) {
        const { TOTAL_POS } = this;
        for (const dirType of rightHalfGomokuDirectionTypes) {
            for (let id = TOTAL_POS - 1; id >= 0; --id)
                handle(id, dirType);
        }
    }
}

class GomokuCountMap {
    context;
    board;
    dirCountMap;
    conShapeCountMap;
    gapShapeCountMap;
    scoreMap;
    constructor(context, board, scoreMap) {
        this.context = context;
        this.board = board;
        this.dirCountMap = new Array(gomokuDirections.length)
            .fill([])
            .map(() => new Uint32Array(context.TOTAL_POS));
        this.conShapeCountMap = new Array(context.TOTAL_PLAYERS)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.gapShapeCountMap = new Array(context.TOTAL_PLAYERS)
            .fill([])
            .map(() => new Array(context.MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]));
        this.scoreMap = scoreMap;
    }
    init() {
        this.dirCountMap.forEach(idMap => idMap.fill(0));
        this.conShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        this.gapShapeCountMap.forEach(maps => maps.forEach(countMap => countMap.fill(0)));
        const { context, board, dirCountMap } = this;
        context.traverseAllDirections((id, dirType) => {
            const player = board[id];
            if (player >= 0) {
                const id2 = context.safeMoveOneStep(id, dirType);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
        });
        context.traverseLeftDirections((id, dirType) => {
            const player = board[id];
            if (player >= 0 && dirCountMap[dirType][id] === 1) {
                const revDirType = dirType ^ 1;
                this.updateConShapeCountMap(id, revDirType, 1);
            }
        });
        context.traverseLeftDirections((id, dirType) => {
            const revDirType = dirType ^ 1;
            if (board[id] < 0) {
                const id0 = context.safeMoveOneStep(id, revDirType);
                if (id0 < 0 || board[id0] < 0)
                    return;
                const player = board[id0];
                const id2 = context.safeMoveOneStep(id, dirType);
                if (id2 < 0 || board[id2] !== player)
                    return;
                const lftCnt = dirCountMap[revDirType][id0];
                const rhtCnt = dirCountMap[dirType][id2];
                const startId = context.fastMove(id, revDirType, lftCnt);
                this.updateGapShapeCountMap(startId, lftCnt, rhtCnt, dirType, 1);
            }
        });
    }
    beforeForward(id) {
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, -1);
        }
    }
    afterForward(id) {
        const { context, board, dirCountMap } = this;
        const player = board[id];
        {
            for (const dirType of gomokuDirectionTypes) {
                const id2 = context.safeMoveOneStep(id, dirType);
                const countMap = dirCountMap[dirType];
                countMap[id] = id2 >= 0 && board[id2] === player ? countMap[id2] + 1 : 1;
            }
            for (const dirType of gomokuDirectionTypes) {
                const revDirType = dirType ^ 1;
                const lftCnt = dirCountMap[revDirType][id];
                const rhtCnt = dirCountMap[dirType][id];
                const countMap = dirCountMap[dirType];
                for (let id2 = id, i = 1; i < lftCnt; ++i) {
                    id2 = context.fastMoveOneStep(id2, revDirType);
                    countMap[id2] += rhtCnt;
                }
            }
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, 1);
        }
    }
    beforeRollback(id) {
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, -1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, -1);
        }
    }
    afterRollback(id, player) {
        const { context, board, dirCountMap } = this;
        {
            for (const dirType of gomokuDirectionTypes) {
                const revDirType = dirType ^ 1;
                const lftCnt = dirCountMap[revDirType][id];
                const rhtCnt = dirCountMap[dirType][id];
                const countMap = dirCountMap[dirType];
                for (let id2 = id, i = 1; i < lftCnt; ++i) {
                    id2 = context.fastMoveOneStep(id2, revDirType);
                    countMap[id2] -= rhtCnt;
                }
            }
            for (const dirType of gomokuDirectionTypes)
                dirCountMap[dirType][id] = 0;
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedConShapeCountMap(id, dirType, 1);
        }
        for (const dirType of leftHalfGomokuDirectionTypes) {
            this.updateRelatedGapShapeCountMap(id, dirType, 1);
        }
    }
    score(currentPlayer, scoreForPlayer) {
        if (this.hasReachedTheLimit(scoreForPlayer))
            return Number.POSITIVE_INFINITY;
        if (this.hasReachedTheLimit(scoreForPlayer ^ 1))
            return Number.NEGATIVE_INFINITY;
        const { context, conShapeCountMap, gapShapeCountMap, scoreMap } = this;
        const conCountMap = conShapeCountMap[scoreForPlayer];
        const gapCountMap = gapShapeCountMap[scoreForPlayer];
        let score = 0;
        for (let cnt = 1; cnt < context.MAX_INLINE; ++cnt) {
            const [a, b, c] = conCountMap[cnt];
            const [x, y, z] = scoreMap.continuously[cnt];
            score += a * x + b * y + c * z;
        }
        for (let cnt = 1; cnt < context.MAX_INLINE; ++cnt) {
            const [a, b, c] = gapCountMap[cnt];
            const [x, y, z] = scoreMap.gap[cnt];
            score += a * x + b * y + c * z;
        }
        if (currentPlayer === scoreForPlayer)
            return score;
        const buffer = Math.random() * context.NEXT_MOVER_MAX_BUFFER;
        return score + score * buffer;
    }
    hasReachedTheLimit(player) {
        const { context, conShapeCountMap } = this;
        return conShapeCountMap[player][context.MAX_INLINE].some(x => x > 0);
    }
    toJSON() {
        const { dirCountMap, conShapeCountMap, gapShapeCountMap } = this;
        return { dirCountMap, conShapeCountMap, gapShapeCountMap };
    }
    updateRelatedConShapeCountMap(centerId, dirType, v) {
        const revDirType = dirType ^ 1;
        const { context, board, dirCountMap } = this;
        const lftMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, revDirType));
        const rhtMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, dirType));
        const THRESHOLD = lftMaxMovableStep + rhtMaxMovableStep + 1;
        let steps = 0;
        let id = context.fastMove(centerId, revDirType, lftMaxMovableStep);
        for (; steps < THRESHOLD;) {
            if (board[id] < 0) {
                steps += 1;
                id = context.fastMoveOneStep(id, dirType);
                continue;
            }
            const rhtCnt = dirCountMap[dirType][id];
            if (steps === 0) {
                const rhtCnt = dirCountMap[dirType][id];
                const id2 = context.fastMove(id, dirType, rhtCnt - 1);
                this.updateConShapeCountMap(id2, revDirType, v);
            }
            else {
                this.updateConShapeCountMap(id, dirType, v);
            }
            id = context.fastMove(id, dirType, rhtCnt);
            steps += rhtCnt;
        }
    }
    updateConShapeCountMap(startId, dirType, v) {
        const player = this.board[startId];
        const cnt = this.dirCountMap[dirType][startId];
        const countOfFreeSide = this.countFreeSide(player, startId, cnt, dirType);
        const normalizedCnt = Math.min(cnt, this.context.MAX_INLINE);
        this.conShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    updateRelatedGapShapeCountMap(centerId, dirType, v) {
        const revDirType = dirType ^ 1;
        const { context, board, dirCountMap } = this;
        const lftMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, revDirType));
        const rhtMaxMovableStep = Math.min(context.MAX_INLINE - 1, context.maxMovableSteps(centerId, dirType));
        const THRESHOLD = lftMaxMovableStep + rhtMaxMovableStep + 1;
        let steps = 0;
        let id = context.fastMove(centerId, revDirType, lftMaxMovableStep);
        for (; steps < THRESHOLD; steps += 1, id = context.fastMoveOneStep(id, dirType)) {
            if (board[id] < 0)
                continue;
            if (steps < 2) {
                const rhtCnt = dirCountMap[dirType][id];
                const id2 = context.fastMove(id, dirType, rhtCnt - 1);
                this.detectGapShape(id2, revDirType, v);
            }
            const lftCnt = dirCountMap[revDirType][id];
            id = context.fastMove(id, revDirType, lftCnt - 1);
            steps -= lftCnt - 1;
            break;
        }
        for (; steps < THRESHOLD;) {
            if (board[id] < 0) {
                id = context.fastMoveOneStep(id, dirType);
                steps += 1;
                continue;
            }
            this.detectGapShape(id, dirType, v);
            const rhtCnt = dirCountMap[dirType][id];
            id = context.fastMove(id, dirType, rhtCnt);
            steps += rhtCnt;
        }
    }
    detectGapShape(id, dirType, v) {
        const { context, board, dirCountMap } = this;
        const lftCnt = dirCountMap[dirType][id];
        if (context.maxMovableSteps(id, dirType) < lftCnt)
            return;
        const id1 = context.fastMove(id, dirType, lftCnt);
        if (board[id1] >= 0 || context.maxMovableSteps(id1, dirType) < 1)
            return;
        const player = board[id];
        const id2 = context.fastMoveOneStep(id1, dirType);
        if (board[id2] !== player)
            return;
        const rhtCnt = dirCountMap[dirType][id2];
        this.updateGapShapeCountMap(id, lftCnt, rhtCnt, dirType, v);
    }
    updateGapShapeCountMap(startId, cnt1, cnt2, dirType, v) {
        const threshold = this.context.MAX_INLINE - 1;
        if (cnt1 >= threshold || cnt2 >= threshold)
            return;
        const player = this.board[startId];
        const cnt = cnt1 + cnt2;
        const countOfFreeSide = this.countFreeSide(player, startId, cnt + 1, dirType);
        const normalizedCnt = Math.min(cnt, this.context.MAX_INLINE);
        this.gapShapeCountMap[player][normalizedCnt][countOfFreeSide] += v;
    }
    countFreeSide(player, startId, cnt, dirType) {
        const { context, board } = this;
        const { MAX_INLINE } = context;
        const revDirType = dirType ^ 1;
        const id0 = context.safeMoveOneStep(startId, revDirType);
        const isFreeSide0 = id0 >= 0 && board[id0] < 0 ? 1 : 0;
        const id2 = context.safeMove(startId, dirType, cnt);
        const isFreeSide2 = id2 >= 0 && board[id2] < 0 ? 1 : 0;
        const countOfFreeSide = isFreeSide0 + isFreeSide2;
        if (countOfFreeSide === 0)
            return 0;
        let maxPossibleCnt = cnt + countOfFreeSide;
        if (maxPossibleCnt < MAX_INLINE) {
            const revPlayer = player ^ 1;
            if (isFreeSide0) {
                const maxMovableSteps = context.maxMovableSteps(id0, revDirType);
                for (let id = id0, step = 0; maxPossibleCnt < MAX_INLINE && step < maxMovableSteps; ++step, ++maxPossibleCnt) {
                    id = context.fastMoveOneStep(id, revDirType);
                    if (board[id] === revPlayer)
                        break;
                }
            }
            if (isFreeSide2) {
                const maxMovableSteps = context.maxMovableSteps(id2, dirType);
                for (let id = id2, step = 0; maxPossibleCnt < MAX_INLINE && step < maxMovableSteps; ++step, ++maxPossibleCnt) {
                    id = context.fastMoveOneStep(id, dirType);
                    if (board[id] === revPlayer)
                        break;
                }
            }
        }
        return maxPossibleCnt < MAX_INLINE ? 0 : countOfFreeSide;
    }
}

class GomokuState {
    context;
    countMap;
    candidateSet;
    board;
    placedCount;
    constructor(context, scoreMap) {
        this.context = context;
        this.board = new Int32Array(context.TOTAL_POS);
        this.countMap = new GomokuCountMap(context, this.board, scoreMap);
        this.candidateSet = new Set();
        this.placedCount = 0;
    }
    init(pieces = []) {
        const { context, board, candidateSet } = this;
        this.placedCount = pieces.length;
        board.fill(-1);
        candidateSet.clear();
        for (const { r, c, p } of pieces) {
            const id = context.idx(r, c);
            board[id] = p;
            candidateSet.delete(id);
            for (const [id2] of context.validNeighbors(id)) {
                if (board[id2] < 0)
                    candidateSet.add(id2);
            }
        }
        this.countMap.init();
    }
    forward(id, player) {
        const { context, board, candidateSet } = this;
        if (id < 0 || board[id] >= 0)
            return;
        this.beforeForward(id);
        {
            this.placedCount += 1;
            board[id] = player;
            candidateSet.delete(id);
            for (const [id2] of context.validNeighbors(id)) {
                if (board[id2] < 0)
                    candidateSet.add(id2);
            }
        }
        this.afterForward(id);
    }
    rollback(id) {
        const { context, board, candidateSet } = this;
        if (id < 0 || board[id] < 0)
            return;
        const player = board[id];
        this.beforeRollback(id);
        {
            this.placedCount -= 1;
            board[id] = -1;
            if (this.hasPlacedNeighbors(id))
                candidateSet.add(id);
            for (const [id2] of context.validNeighbors(id)) {
                if (board[id2] >= 0 || !this.hasPlacedNeighbors(id2)) {
                    candidateSet.delete(id2);
                }
            }
        }
        this.afterRollback(id, player);
    }
    expand(currentPlayer, scoreForPlayer) {
        const { board, candidateSet } = this;
        const candidates = [];
        for (const id of candidateSet) {
            if (board[id] >= 0)
                continue;
            candidates.push({ id, score: 0 });
        }
        for (const candidate of candidates) {
            this.forward(candidate.id, currentPlayer);
            candidate.score = this.score(currentPlayer, scoreForPlayer);
            this.rollback(candidate.id);
        }
        return candidates;
    }
    score(currentPlayer, scoreForPlayer) {
        return (this.countMap.score(currentPlayer, scoreForPlayer) -
            this.countMap.score(currentPlayer, scoreForPlayer ^ 1));
    }
    isFinal() {
        const { context, countMap, placedCount } = this;
        if (placedCount === context.TOTAL_POS)
            return true;
        for (let player = 0; player < context.TOTAL_PLAYERS; ++player) {
            if (countMap.hasReachedTheLimit(player))
                return true;
        }
        return false;
    }
    randomMove() {
        const { context, board, candidateSet } = this;
        for (const id of candidateSet) {
            for (const [id2] of context.validNeighbors(id)) {
                if (board[id2] < 0)
                    return id2;
            }
        }
        return -1;
    }
    hasPlacedNeighbors(id) {
        const { context, board } = this;
        for (const [id2] of context.validNeighbors(id)) {
            if (board[id2] >= 0)
                return true;
        }
        return false;
    }
    beforeForward(id) {
        this.countMap.beforeForward(id);
    }
    afterForward(id) {
        this.countMap.afterForward(id);
    }
    beforeRollback(id) {
        this.countMap.beforeRollback(id);
    }
    afterRollback(id, player) {
        this.countMap.afterRollback(id, player);
    }
}

class GomokuStateCompressor {
    MOVE_STEP_BIT_BASE;
    MOVE_STEP_MASK;
    MOVE_STEP_MINIMUM_FLAG;
    INITIAL_STATE = 1n;
    constructor(TOTAL_POS) {
        let COMPRESS_MOVE_BIT_BASE = 1n;
        while (1n << COMPRESS_MOVE_BIT_BASE < TOTAL_POS)
            COMPRESS_MOVE_BIT_BASE += 1n;
        this.MOVE_STEP_BIT_BASE = COMPRESS_MOVE_BIT_BASE;
        this.MOVE_STEP_MASK = (1n << COMPRESS_MOVE_BIT_BASE) - 1n;
        this.MOVE_STEP_MINIMUM_FLAG = 1n << COMPRESS_MOVE_BIT_BASE;
    }
    compress(countInState, state, newMoveStepId) {
        if (countInState === 0)
            return newMoveStepId + this.MOVE_STEP_MINIMUM_FLAG;
        if (countInState === 1)
            return (state << this.MOVE_STEP_BIT_BASE) | newMoveStepId;
        const rhtCnt = countInState >> 1;
        const RIGHT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(rhtCnt);
        const rht = state & ((1n << RIGHT_BITS) - 1n);
        if (countInState & 1) {
            const rhtMoveIds = this.decompressMoveIds(rhtCnt, rht, newMoveStepId);
            const newRht = this.compressMoveIds(rhtMoveIds);
            return ((state ^ rht) << this.MOVE_STEP_BIT_BASE) | newRht;
        }
        else {
            const lftCnt = countInState - rhtCnt;
            const lft = state >> RIGHT_BITS;
            const lftMoveIds = this.decompressMoveIds(lftCnt, lft, newMoveStepId);
            const NEW_LEFT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(lftCnt + 1);
            const newLft = this.compressMoveIds(lftMoveIds) | (1n << NEW_LEFT_BITS);
            return (newLft << RIGHT_BITS) | rht;
        }
    }
    decompressMoveIds(cnt, state, newMoveStepId) {
        const { MOVE_STEP_BIT_BASE, MOVE_STEP_MASK } = this;
        const moveIds = [];
        let i = 0;
        let v = state;
        let id = 0n;
        for (; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            if (id < newMoveStepId)
                break;
            moveIds.push(id);
        }
        moveIds.push(newMoveStepId);
        if (i < cnt)
            moveIds.push(id);
        for (i += 1; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            moveIds.push(id);
        }
        return moveIds.reverse();
    }
    compressMoveIds(moveIds) {
        const { MOVE_STEP_BIT_BASE } = this;
        let result = 0n;
        for (const id of moveIds) {
            result = (result << MOVE_STEP_BIT_BASE) | id;
        }
        return result;
    }
}

const createScoreMap = (MAX_INLINE) => {
    const continuously = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    const gap = new Array(MAX_INLINE + 1).fill([]).map(() => [0, 0, 0]);
    let maxValue = 2 ** 30;
    const QUARTER_OF_MAX_VALUE = maxValue / 4;
    let cnt = MAX_INLINE;
    continuously[cnt] = [maxValue, maxValue, maxValue];
    gap[cnt] = [QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE];
    if (MAX_INLINE > 1) {
        cnt -= 1;
        continuously[cnt] = [0, QUARTER_OF_MAX_VALUE, maxValue / 2];
        gap[cnt] = [QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE, QUARTER_OF_MAX_VALUE];
    }
    maxValue /= 8;
    for (; cnt > 0; --cnt, maxValue /= 8) {
        continuously[cnt] = [0, maxValue / 4, maxValue];
        gap[cnt] = [0, maxValue / 8, maxValue / 4];
    }
    return { continuously, gap };
};

class GomokuSolution {
    MAX_DEPTH;
    context;
    state;
    stateCompressor;
    stateCache;
    scoreForPlayer;
    bestMoveId;
    constructor(MAX_ROW, MAX_COL, MAX_INLINE = 5, MAX_DEPTH = 3, NEXT_MOVER_MAX_BUFFER, scoreMap) {
        const context = new GomokuContext(MAX_ROW, MAX_COL, MAX_INLINE, NEXT_MOVER_MAX_BUFFER);
        const _scoreMap = scoreMap ?? createScoreMap(context.MAX_INLINE);
        this.MAX_DEPTH = Math.max(1, Math.round(MAX_DEPTH));
        this.context = context;
        this.state = new GomokuState(context, _scoreMap);
        this.stateCompressor = new GomokuStateCompressor(BigInt(context.TOTAL_POS));
        this.stateCache = new Map();
        this.scoreForPlayer = -1;
        this.bestMoveId = -1;
    }
    init(pieces) {
        this.state.init(pieces);
        this.stateCache.clear();
    }
    forward(r, c, p) {
        const id = this.context.idx(r, c);
        this.state.forward(id, p);
    }
    rollback(r, c) {
        const id = this.context.idx(r, c);
        this.state.rollback(id);
    }
    minmaxMatch(currentPlayer) {
        if (this.state.isFinal())
            return [-1, -1];
        this.stateCache.clear();
        this.scoreForPlayer = currentPlayer;
        this.bestMoveId = -1;
        this.alphaBeta(currentPlayer, Number.NEGATIVE_INFINITY, Number.POSITIVE_INFINITY, 0, this.stateCompressor.INITIAL_STATE, 0);
        const bestMoveId = this.bestMoveId < 0 ? this.state.randomMove() : this.bestMoveId;
        const [r, c] = this.context.revIdx(bestMoveId);
        return [r, c];
    }
    alphaBeta(player, alpha, beta, cur, prevState, stateScore) {
        const { state, stateCompressor, stateCache, scoreForPlayer } = this;
        if (cur === this.MAX_DEPTH || state.isFinal())
            return stateScore;
        const candidates = state.expand(player, scoreForPlayer);
        if (player === scoreForPlayer) {
            candidates.sort((x, y) => y.score - x.score);
        }
        else {
            candidates.sort((x, y) => x.score - y.score);
        }
        for (const { id, score } of candidates) {
            const nextState = stateCompressor.compress(cur, prevState, BigInt(id));
            let gamma = stateCache.get(nextState);
            if (gamma === undefined) {
                state.forward(id, player);
                gamma = this.alphaBeta(player ^ 1, alpha, beta, cur + 1, nextState, score);
                state.rollback(id);
                stateCache.set(nextState, gamma);
            }
            if (cur === 0 && alpha < gamma)
                this.bestMoveId = id;
            if (player === scoreForPlayer) {
                if (alpha < gamma)
                    alpha = gamma;
            }
            else {
                if (beta > gamma)
                    beta = gamma;
            }
            if (beta <= alpha)
                break;
        }
        return player === scoreForPlayer ? alpha : beta;
    }
}

exports.GomokuContext = GomokuContext;
exports.GomokuCountMap = GomokuCountMap;
exports.GomokuSolution = GomokuSolution;
exports.GomokuState = GomokuState;
exports.GomokuStateCompressor = GomokuStateCompressor;
exports.createScoreMap = createScoreMap;
exports.gomokuDirectionTypes = gomokuDirectionTypes;
exports.gomokuDirections = gomokuDirections;
exports.leftHalfGomokuDirectionTypes = leftHalfGomokuDirectionTypes;
exports.rightHalfGomokuDirectionTypes = rightHalfGomokuDirectionTypes;
